import javax.swing.*;
import javax.swing.tree.*;
import javax.swing.event.*;
import java.awt.*;
import java.io.*;

import CH.psi.midas.*;

public class OdbTree extends javax.swing.JTree
{
    // Member variables
    Experiment experiment = null;
    
	public OdbTree()
	{
        // Create JTree itself
        super((TreeModel)null);
        
        // Use horizontal and vertical lines
        putClientProperty("JTree.lineStyle", "Angled");

		//{{INIT_CONTROLS
		setLayout(null);
		Insets ins = getInsets();
		setSize(ins.left + ins.right + 430,ins.top + ins.bottom + 270);
		//}}

		//{{REGISTER_LISTENERS
		//}}
	}

	//{{DECLARE_CONTROLS
	//}}
	
  public void connect(Experiment experiment, String path) throws IOException,OdbKeyNotFoundException {
    this.experiment = experiment;
 
    // Create the first node
    OdbKey rootKey = new OdbKey(experiment, path);
    OdbTreeNode rootNode = new OdbTreeNode(null, rootKey);
    
    // populate root node with first level of subdirectories
    rootNode.populateDirectories();
    setModel(new DefaultTreeModel(rootNode));
    
    // listen for Tree Selection Events
    addTreeExpansionListener(new TreeExpansionHandler());
  }

  // Inner class that represents a ODB node
  protected static class OdbTreeNode extends DefaultMutableTreeNode {
    public OdbTreeNode(String parent, OdbKey odbKey) throws
             IOException,OdbKeyNotFoundException {
      
      this.odbKey = odbKey;

      // construct full name
      if (parent == null)
        fullName = odbKey.getName();
      else {
        if (parent.equals("/"))
          fullName = "/"+odbKey.getName();
        else
          fullName = parent +"/"+odbKey.getName();
      }
    }

  // Override isLeaf to check whether this is a directory
  public boolean isLeaf() {
    return odbKey.getType() != OdbKey.TID_KEY;
  }
  
  // Override getAllowsChildren to check whether this is a direcotry
  public boolean getAllowsChildren() {
    return odbKey.getType() == OdbKey.TID_KEY;
  }
  
  // For display purposes, we return our onw name
  public String toString() {
    return odbKey.getName();
  }
  
  // scan ODB contents and populate tree
  private void populateDirectories() {
    
    // first remove all children
    removeAllChildren();
      
    // enumerate keys
    try {
      for (int i=0 ; ; i++) {
        OdbKey sk = odbKey.enumSubKey(i);
        if (sk == null)
        break;
            
        OdbTreeNode node = new OdbTreeNode(fullName, sk);
        this.add(node);
          
        // create dummy child if key with subkeys
        if (sk.getType() == OdbKey.TID_KEY && sk.getNumValues()>0) {
          DefaultMutableTreeNode dummy = new DefaultMutableTreeNode();
          node.add(dummy);
          //node.populateDirectories(false);
        }
      }
    } catch (Exception e) {}
  }
  
  protected String fullName;
  protected OdbKey odbKey;
  }
  
  public OdbKey getPathKey(TreePath path) {
    Object o = path.getLastPathComponent();
    if (o instanceof OdbTreeNode)
      return ((OdbTreeNode)o).odbKey;

    System.out.println("Unknown node "+o);
    return null;
  }
  
  // inner class that handles Tree Expansion Events
  protected class TreeExpansionHandler implements TreeExpansionListener {
    public void treeExpanded(TreeExpansionEvent evt) {
      TreePath path = evt.getPath();
      JTree tree = (JTree)evt.getSource();
      
      // populate last component of the path
      OdbTreeNode node = (OdbTreeNode)path.getLastPathComponent();
      node.populateDirectories();
      ((DefaultTreeModel)tree.getModel()).nodeStructureChanged(node);
    }
    
    public void treeCollapsed(TreeExpansionEvent evt) {
    }
  }
}